"""
Session and session state management for standman.

This module defines the Session lifecycle and provides a read-only
SessionState interface to check whether a session is active and if
any error has occurred.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from threading import Lock

class SessionState(ABC):
    """Read-only interface for observing a session's state."""
    __slots__ = ()
    @property
    @abstractmethod
    def ok(self) -> bool:
        """Whether the session is still active."""
    
    @property
    @abstractmethod
    def error(self) -> Exception | None:
        """Error that caused the session to stop, or None if none."""


class Session:
    """Internal session controller.

    Tracks active state and error status. Provides
    a SessionState reader for external observers.
    """
    
    __slots__ = ('_lock', '_active', '_error')
    def __init__(self):
        self._lock = Lock()
        self._active = True
        self._error = None
    
    @property
    def ok(self) -> bool:
        """Whether the session is still active."""
        with self._lock:
            return self._active
    
    @property
    def error(self) -> Exception | None:
        """Return the first error recorded, if any."""
        with self._lock:
            return self._error

    def set_error(self, exc: Exception) -> None:
        """Mark the session as failed with the given exception."""
        with self._lock:
            if self._error is None:
                self._error = exc
            self._active = False
    
    def get_state_reader(self):
        """Return a read-only view of the session state."""

        outer = self

        class _SessionState(SessionState):
            @property
            def ok(self) -> bool:
                return outer.ok
            
            @property
            def error(self) -> Exception | None:
                return outer.error
        
        return _SessionState()







from __future__ import annotations

import enum
from typing import Callable


class ProcessObserver:
    __slots__ = ('_conditions', '_global_violation', '_global_fail_reason', '_global_exception',
                 '_local_violation', '_observations', '_violation_handlers', '_exception_handler')

    def __init__(self, conditions: dict[str, Callable[..., bool]]):
        self._conditions = conditions

        self._global_violation = False
        self._global_fail_reason = ''
        self._global_exception = None

        self._local_violation = False

        self._observations = {tag: Observation() for tag in conditions.keys()}

        self._violation_handlers = {}

        self._exception_handler = None


    def reset_observations(self) -> None:
        self._global_violation = False
        self._global_fail_reason = ''
        self._global_exception = None

        self._local_violation = False

        self._observations = {tag: Observation() for tag in self._conditions.keys()}

    
    def listen(self, tag: str, *args, **kwargs) -> None:
        try:
            if tag not in self._observations:
                if not self._global_violation:
                    self._global_violation = True
                    self._global_fail_reason = f"wrong tag '{tag}'"
                return
            
            observation = self._observations[tag]
            condition = self._conditions[tag]
            pass_ = False
            try:
                pass_ = condition(*args, **kwargs)
            except Exception as e:
                self._local_violation = True
                if not observation.violation:
                    observation.violation = True
                    observation.first_violation_at = observation.count
                    observation.fail_condition = condition
                    observation.fail_reason = f'exception at {tag} at {observation.count}th attempt'
                    observation.exc = e
                    self._call_exception_handler(tag, ExceptionKind.ON_CONDITION, observation, e)
                self._call_violation_handler(tag, observation)


            if not pass_:
                self._local_violation = True
                if not observation.violation:
                    observation.violation = True
                    observation.first_violation_at = observation.count
                    observation.fail_condition = condition
                    observation.fail_reason = 'condition violation'
                self._call_violation_handler(tag, observation)
            
            observation.count += 1
        except Exception as e:
            # overrides all global violations
            self._global_violation = True
            self._global_fail_reason = "internal error"
            self._global_exception = e
            self._call_exception_handler(tag, ExceptionKind.ON_INTERNAL, None, e)


    def _call_violation_handler(self, tag, observation):
        if tag in self._violation_handlers:
            try:
                self._violation_handlers[tag](observation)
            except Exception as e:
                self._call_exception_handler(tag, ExceptionKind.ON_VIOLATION, observation, e)
                pass
    
    def _call_exception_handler(self, tag, kind, observation, e):
        if self._exception_handler:
            try:
                self._exception_handler(tag, kind, observation, e)
            except Exception:
                pass
    
    @property
    def violation(self):
        return self._global_violation or self._local_violation
    
    @property
    def global_violation(self):
        return self._global_violation
    
    @property
    def local_violation(self):
        return self._local_violation
    
    @property
    def global_fail_reason(self) -> str:
        return self._global_fail_reason
    
    @property
    def global_exception(self) -> Exception | None:
        return self._global_exception
    
    def get_all(self) -> dict[str, Observation]:
        return {k: v for k, v in self._observations.items()}
    
    def get_violated(self) -> dict[str, Observation]:
        return {k: v for k, v in self._observations.items() if v.violation}
    
    def get_compliant(self) -> dict[str, Observation]:
        return {k: v for k, v in self._observations.items() if not v.violation}
    
    def get_unevaluated(self) -> dict[str, Observation]:
        return {k: v for k, v in self._observations.items() if v.count == 0}

    def set_violation_handler(self, tag: str, fn: Callable[[Observation], None]) -> None:
        if tag not in self._conditions:
            raise ValueError(f"Condition '{tag}' is not defined")
        self._violation_handlers[tag] = fn
    
    def set_exception_handler(self, fn: Callable[[str, ExceptionKind, Observation | None, Exception], None]) -> None:
        self._exception_handler = fn

    def get_stat(self, tag: str) -> ConditionStat:
        observation = self._observations[tag]
        stat = ConditionStat(observation.count, observation.violation, observation.first_violation_at)
        return stat


class Observation:
    '''Detailed observation results by condition.'''

    __slots__ = ('count', 'violation', 'first_violation_at', 'exc', 'fail_condition', 'fail_reason')
    def __init__(self):
        self.count: int = 0
        self.violation: bool = False
        self.first_violation_at: int = -1
        self.exc: Exception | None = None
        self.fail_condition: Callable[..., bool] | None = None
        self.fail_reason: str = ''


class ConditionStat:
    '''Represents a simplified statistical view for a specific condition.'''

    __slots__ = ('_count', '_violation', '_first_violation_at')
    def __init__(self, count: int, violation: bool, first_violation_at: int):
        self._count = count
        self._violation = violation
        self._first_violation_at = first_violation_at
    
    @property
    def count(self) -> int:
        return self._count
    
    @property
    def violation(self) -> bool:
        return self._violation
    
    @property
    def first_violation_at(self) -> int:
        return self._first_violation_at


class ExceptionKind(enum.Enum):
    ON_CONDITION = 'Exception raised on condition.'
    ON_VIOLATION = 'Exception raised on violation handler'
    ON_INTERNAL = 'Exception raised on internal'




"""
Custom exception definitions for standman.

This module provides framework-specific errors used in session
and port management.
"""

class OccupiedError(Exception):
    """Raised when a Port is already occupied by another session."""
    pass

class DeniedError(Exception):
    """Raised when a connection is denied by the policy or Port."""
    pass




"""
standman package initializer.

This module defines the public API of standman by re-exporting
interfaces from internal modules. Users should import symbols
from here instead of directly accessing submodules.

Exports:
    - SessionPolicy, create_session_policy : Manage Ports and Sessions
    - Port                                : Interface for sending data
    - SessionState                        : Read-only session state
    - SendFunction, ListenFunction        : Protocols for callbacks
    - DeniedError, OccupiedError          : Exceptions for connection control
    - __version__                         : Package version

Design note:
    Asynchronous interfaces are intentionally not supported.
    Listener implementations may take any form, but the design
    ensures that sending side code is never affected by exceptions,
    serialization, or concurrency side effects introduced here.

See also:
    The `example()` function in this module demonstrates
    a minimal working usage of SessionPolicy, Port, and session.

"""

from .policy import SessionPolicy, create_session_policy
from .port import Port
from .session import SessionState
from .protocols import SendFunction, ListenFunction
from .exceptions import DeniedError, OccupiedError

from .observer import ProcessObserver

__version__ = '1.0.3'

__all__ = (
    'SessionPolicy', 'create_session_policy',
    'Port',
    'SessionState',
    'SendFunction', 'ListenFunction',
    'DeniedError', 'OccupiedError',
    '__version__')

def example():
    policy = create_session_policy()
    port = policy.create_port()

    # Implementation-side function
    def add(a, b):
        port.send("add", a, b)
        return a + b

    # Listener function
    def listener(tag, *args, **kwargs):
        print(f"Received: {tag}, args={args}, kwargs={kwargs}")

    # Run with session
    with policy.session(listener, port) as state:
        result = add(2, 3)
        assert state.ok
        assert state.error is None
        print("Result:", result)

    # Output:
    # Received: add, args=(2, 3), kwargs={}
    # Result: 5


def example_with_observer():

    def create_weather_sensor(port):
        """Weather sensor
        Specification:
            temp < 0        -> "Freezing" + send("freezing")
            0 <= temp <= 30 -> "Normal"   + send("normal")
            temp > 30       -> "Hot"      + send("hot")
        """
        def check_weather(temp: int) -> str:
            # If there is a bug here, it will be detected by the test
            if temp <= 0:   # ← Common place to inject a bug
                port.send("freezing", temp)
                return "Freezing"
            elif temp <= 30:
                port.send("normal", temp)
                return "Normal"
            else:
                port.send("hot", temp)
                return "Hot"
        return check_weather

    policy = create_session_policy()
    port = policy.create_port()

    # Define expected conditions according to the specification
    conditions = {
        "freezing": lambda t: t < 0,
        "normal":   lambda t: 0 <= t <= 30,
        "hot":      lambda t: t > 30,
    }
    observer = ProcessObserver(conditions)
    check_weather = create_weather_sensor(port)

    with policy.session(observer.listen, port) as state:
        # Test coverage for all three branches
        for i in (-5, 0, 31):
            check_weather(i)
            if not state.ok:
                raise AssertionError(f"observation failed on '{i}'")

        # Verify that the Observer did not detect any specification violations
        if observer.violation:
            details = []
            for tag, obs in observer.get_violated().items():
                details.append(
                    f"[{tag}] reason={obs.fail_reason}, "
                    f"count={obs.count}, first_violation_at={obs.first_violation_at}"
                )
            raise AssertionError("Observer detected violations:\n" + "\n".join(details))
    
    print("All checks passed!")


if __name__ == "__main__":
    example()



"""
SessionPolicy implementation for standman.

This module provides the SessionPolicy interface and its factory
function create_session_policy(). A SessionPolicy manages Port
creation and session lifecycle, controlling how senders and
listeners are linked together.


Notes
-----

The implementation of this module is based on the `Verbose Pattern`,
which is intended to structure the implementation.

Verbose Pattern:
    https://github.com/minoru-jp/design-notes/blob/main/verbose-pattern/README.md

"""


from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field

from threading import Lock
from typing import Callable, ContextManager, Protocol, cast
from contextlib import contextmanager

from .port import Port, _create_port, _create_noop_port, _create_port_role
from .port import _RoleTOC as _PortRoleTOC
from .protocols import ListenFunction, SendFunction
from .session import Session, SessionState
from .exceptions import DeniedError

class SessionPolicy(ABC):
    """Management interface for creating Ports and sessions."""

    @abstractmethod
    def create_port(self) -> Port:
        """Create a Port available for connections."""
    @abstractmethod
    def create_noop_port(self) -> Port:
        """Create a Port that rejects connections."""

    @abstractmethod
    def session(self, listener: ListenFunction, target: Port) -> ContextManager[SessionState]:
        """
        Establish a connection to the specified Port.

        Args:
            listener:
                Handler for inputs sent to the target.
            target:
                The Port to connect to.

        Raises:
            TypeError:
                Raised if the target is not an instance of Port.
            OccupiedError:
                Raised if another session has already started for the target.
            DeniedError:
                Raised if the Port or SessionPolicy is configured to reject connections.
            RuntimeError:
                Raised if an unexpected internal state is encountered.
                This exception should not normally occur in regular usage.

        Returns:
            A context manager that controls the start and end of the session.
        """

# _*TOC: TOC = Table of Content

class _ConstantTOC(Protocol):
    """Internal structure: constants"""
    SENTINELS: dict[str, Callable]

class _StateTOC(Protocol):
    """Internal structure: state variables"""
    local_lock: Lock

    session_map: dict[Port, Session]

    mess_validator: tuple[Callable]

    entry_permit: object
    control_permit: object

class _KernelTOC(Protocol):
    def create_port(self, bridge: _PortBridgeTOC) -> _PortRoleTOC | Port:
        ...
    
    def create_noop_port(self, bridge: _PortBridgeTOC) -> Port:
        ...

class _CoreTOC(Protocol):
    """Internal structure: core functions"""
    def register_session(self, listen: ListenFunction, target: Port) -> Session:
        ...
    
    def unregister_session(self, target: Port) -> None:
        ...

    def create_port(self) -> Port:
        ...
    
    def create_noop_port(self) -> Port:
        ...
    
    def session(self, listen: ListenFunction, target: Port) -> ContextManager[SessionState]:
        ...

class _PortBridgeTOC(Protocol):
    """Internal structure: delegation functions for Port"""
    def get_session(self, port: Port) -> Session | None:
        ...

    def get_entry_permit(self):
        ...
    
    def get_control_permit(self) -> object:
        ...
    
    def get_message_validator(self) -> SendFunction:
        ...
    
class _RoleTOC(Protocol):
    """Internal structure: accessors to role-specific interfaces"""
    constant: _ConstantTOC
    state: _StateTOC
    kernel: _KernelTOC
    core: _CoreTOC
    port_bridge: _PortBridgeTOC
    interface: SessionPolicy


def _create_session_policy_role(
        *,
        block_port: bool = False,
        message_validator: SendFunction | None = None
) -> _RoleTOC:

    class _Constant(_ConstantTOC):
        __slots__ = ()
        SENTINELS = {"DEFAULT_MESSAGE_VALIDATOR": lambda tag, *a, **kw: None}
    
    constant = _Constant()


    class _State(_StateTOC):
        __slots__ = ()

        local_lock: Lock = Lock()

        session_map: dict[Port, Session] = {}

        entry_permit: object = object()
        control_permit: object = object()

        mess_validator = (message_validator if message_validator else constant.SENTINELS["DEFAULT_MESSAGE_VALIDATOR"],)

    state = _State()

    class _Kernel(_KernelTOC):
        def create_port(self, bridge: _PortBridgeTOC) -> _PortRoleTOC | Port:
            if not block_port:
                return _create_port_role(port_bridge)
            else:
                return _create_noop_port(port_bridge)
        
        def create_noop_port(self, bridge: _PortBridgeTOC) -> Port:
            return _create_noop_port(port_bridge)
    
    kernel = _Kernel()

    class _Core(_CoreTOC):
        
        def register_session(self, listen: ListenFunction, target: Port) -> Session:
            with state.local_lock:

                target._set_listen_func(state.control_permit, listen)
                
                if target in state.session_map:
                    target._remove_listen_func(state.control_permit)
                    raise RuntimeError("Internal error: A session for this target is already registered.")
                
                session = Session()
                state.session_map[target] = session
            
            return session
        
        def unregister_session(self, target: Port) -> None:
            with state.local_lock:
                try:
                    target._remove_listen_func(state.control_permit)
                    state.session_map.pop(target)
                except KeyError as e:
                    raise RuntimeError(f"Internal error: Session not found") from e

        def create_port(self) -> Port:
            obj = kernel.create_port(port_bridge)
            return obj.interface if not isinstance(obj, Port) else obj
        
        def create_noop_port(self) -> Port:
            return kernel.create_noop_port(port_bridge)
        
        def session(self, listen: ListenFunction, target: Port) -> ContextManager[SessionState]:
            if not isinstance(target, Port):
                raise TypeError(f"target must be Port but receives '{type(target)}'")
            
            if target._get_entry_permit() is not state.entry_permit:
                raise DeniedError("target is not created by this policy.")
            
            @contextmanager
            def session_context():
                session = core.register_session(listen, target)
                yield session.get_state_reader()
                core.unregister_session(target)
                
            return session_context()
    
    core = _Core()

    class _PortBridge(_PortBridgeTOC):
        def get_session(self, port: Port) -> Session | None:
            with state.local_lock:
                return state.session_map.get(port, None)
        
        def get_entry_permit(self):
            return state.entry_permit
        
        def get_control_permit(self) -> object:
            return state.control_permit
        
        def get_message_validator(self):
            return state.mess_validator[0]
    
    port_bridge = _PortBridge()

    class _Interface(SessionPolicy):
        def create_port(self) -> Port:
            return core.create_port()
        
        def create_noop_port(self) -> Port:
            return core.create_noop_port()
        
        def session(self, listener: ListenFunction, target: Port) -> ContextManager[SessionState]:
            return core.session(listener, target)

    interface = _Interface()


    @dataclass(slots = True)
    class _Role(_RoleTOC):
        constant: _ConstantTOC
        state: _StateTOC
        kernel: _KernelTOC
        core: _CoreTOC
        port_bridge: _PortBridgeTOC
        interface: SessionPolicy

    return _Role(
        constant = constant,
        state = state,
        kernel = kernel,
        core = core,
        port_bridge = port_bridge,
        interface = interface)


def create_session_policy(
        *,
        block_port = False,
        message_validator: SendFunction | None = None
) -> SessionPolicy:
    """
    Create a SessionPolicy interface.

    Args:
        block_port:
            If True, connections to Ports created by this policy will be rejected.
        message_validator:
            A function to validate messages before Port.send() is called.
            By raising an exception, the send can be rejected.
            Exceptions raised by this validator are not propagated from Port.send(),
            and instead the session ends silently. Such errors can be detected
            through the receiver's SessionState.

    Returns:
        SessionPolicy:
            An interface for creating Ports and establishing sessions.
    """
    role = _create_session_policy_role(
        block_port = block_port,
        message_validator= message_validator)
    return role.interface






"""
Protocols for callback functions in standman.

Defines typing contracts for senders and listeners used
in Port and Session communication.
"""

from typing import Protocol


class SendFunction(Protocol):
    def __call__(self, tag: str, *args, **kwargs) -> None:
        """A callable used by implementation side to send messages.

        Args:
            tag (str): Identifier for the message.
            *args: Arbitrary positional arguments for message data.
            **kwargs: Arbitrary keyword arguments for message data.
        """

class ListenFunction(Protocol):
    def __call__(self, tag: str, *args, **kwargs) -> None:
        """A callable used by receiver side to handle incoming messages.

        Args:
            tag (str): Identifier for the message.
            *args: Arbitrary positional arguments passed from sender.
            **kwargs: Arbitrary keyword arguments passed from sender.
        """




"""
Port interface and factory functions for standman.

Ports are used on the implementation (sender) side to send
information to observers. Ports are created via SessionPolicy
and are responsible for linking sender functions to listener
callbacks in a fail-silent manner.


Design note:
    Ports are designed not to introduce unintended side effects
    into the implementation: exceptions never propagate back to
    the sender, and send() does not enforce serialization that
    could mask concurrency issues.

    Registration and removal of listeners are thread-safe, but this is
    provided for potential future changes and extensions. At present,
    only single-threaded usage is assumed.
    In contrast, Port.send() is always thread-unsafe, and this will never
    change, even with future modifications or extensions.

Scope of Port usage:
    Ports can be defined with different scopes depending on the use case:

    * Module-level:
        A Port shared across multiple functions within a module.
        Example:
            port = create_port()
            def func1():
                port.send("example", "func1")
            def func2():
                port.send("example", "func2")
    
    * Function-level:
        Each function holds its own Port.
        Example:
            def sender():
                port = sender._port
                port.send("example", "sender")
            sender.port = create_port()

    * Class-level:
        A Port shared across all instances of a class.
        Example:
            class Foo:
                port = create_port()
                def method(self):
                    Foo.port.send("example", "Foo.method")

            class Bar(Foo):
                def method(self):
                    super().method()
                    Foo.port.send("example", "Bar.method")
    
    * Instance-level:
        Each object instance has its own Port.
        Example:
            class Foo:
                def __init__(self):
                    self.port = create_port()
                
                def method(self):
                    self.port.send("example", id(self))
            
            class Bar(Foo):
                def __init__(self):
                    super().__init__()
                
                def method(self):
                    super().method()
                    self.port.send("example", id(self))

    Note:
        Apply encapsulation and access restrictions to Ports as needed.

"""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from threading import Lock
from typing import TYPE_CHECKING, Protocol

from .protocols import ListenFunction
from .exceptions import OccupiedError, DeniedError

if TYPE_CHECKING:
    from .policy import _PortBridgeTOC

class Port(ABC):
    """Interface for sending information from the implementation side.

    Notes:
        - Thread-safe for listener registration/removal.
        - Thread-unsafe for send() execution to avoid unnecessary
          serialization in concurrent implementations.
    """
    __slots__ = ()
    @abstractmethod
    def send(self, tag: str, *args, **kwargs) -> None:
        """Send arbitrary information to the registered listener.

        Does nothing if no listener is registered.
        Never raises an exception to the sender.

        Note:
            This method is thread-unsafe.
            This is intentional, not for performance optimization,
            but to avoid introducing unintended serialization
            into the implementation through the use of port.send().
            Otherwise, concurrency issues or design flaws in the
            implementation could be hidden by forced serialization.
        
        Args:
            tag (str): Identifier string for the message.
            *args: Arbitrary positional arguments.
            **kwargs: Arbitrary keyword arguments.
        """

    @abstractmethod
    def _set_listen_func(self, key: object, listen: ListenFunction) -> None:
        """Register a listener callback (internal use only)."""

    @abstractmethod
    def _remove_listen_func(self, key: object) -> None:
        """Remove a listener callback (internal use only)."""

    @abstractmethod
    def _get_entry_permit(self) -> object:
        """Return the identifier of the SessionPolicy that created this Port."""


class _StateTOC(Protocol):
    lock: Lock
    listen_func: ListenFunction | None
    error: Exception | None


class _RoleTOC(Protocol):
    state: _StateTOC
    interface: Port


def _create_port_role(bridge: _PortBridgeTOC) -> _RoleTOC:

    @dataclass(slots = True)
    class _State(_StateTOC):
        lock: Lock = field(default_factory = Lock)
        listen_func: ListenFunction | None = field(default = None)
        error: Exception | None = field(default = None)
    
    state = _State()

    class _Interface(Port):
        __slots__ = ()
        
        def send(self, tag: str, *args, **kwargs) -> None:
            try:
                listen_func = state.listen_func
                error = state.error
                if listen_func and not error:
                    bridge.get_message_validator()(tag, *args, **kwargs)
                    listen_func(tag, *args, **kwargs)
            except Exception as e:
                state.error = e
                session = bridge.get_session(self)
                if session is not None:
                    session.set_error(e)
            finally:
                return None
        
        def _set_listen_func(self, key: object, listen: ListenFunction) -> None:
            with state.lock:
                # If an error has already occurred, do nothing instead of raising OccupiedError.
                if state.error:
                    return
                if key is not bridge.get_control_permit():
                    raise PermissionError("Verification failed")
                if state.listen_func is not None:
                    raise OccupiedError("Port is already occupied by another session.")
                state.listen_func = listen
        
        def _remove_listen_func(self, key: object) -> None:
            with state.lock:
                if key is not bridge.get_control_permit():
                    raise PermissionError("Verification failed")
                state.listen_func = None
                state.error = None
                
        
        def _get_entry_permit(self) -> object:
            return bridge.get_entry_permit()

    interface = _Interface()

    @dataclass(slots = True)
    class _Role(_RoleTOC):
        state: _StateTOC
        interface: Port

    return _Role(state = state, interface = interface)


def _create_port(bridge: _PortBridgeTOC) -> Port:
    """Factory: create a functional Port linked to a SessionPolicy."""
    role = _create_port_role(bridge)
    return role.interface


def _create_noop_port(bridge: _PortBridgeTOC) -> Port:
    """Factory: create a no-op Port that denies all connections."""

    class _Interface(Port):
        __slots__ = ()
        def send(self, tag: str, *args, **kwargs) -> None:
            pass
        
        def _set_listen_func(self, key: object, listen: ListenFunction) -> None:
            if key is not bridge.get_control_permit():
                raise PermissionError("Verification failed")
            raise DeniedError("Connection is denied by the policy.")
        
        def _remove_listen_func(self, key: object) -> None:
            if key is not bridge.get_control_permit():
                raise PermissionError("Verification failed")
        
        def _get_entry_permit(self) -> object:
            return bridge.get_entry_permit()

    interface = _Interface()

    return interface

